// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause

#ifndef vtkVTK_USE_SCALED_SOA_ARRAYS_h
#define vtkVTK_USE_SCALED_SOA_ARRAYS_h

// defined if VTK provides the vtkScaledSOADataArray classes
/* #undef VTK_USE_SCALED_SOA_ARRAYS */

#endif // vtkVTK_USE_SCALED_SOA_ARRAYS_h
