// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause
// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause
// This file is autogenerated by vtkCreateArrayDispatchImplicitList.cmake.
// Do not edit this file. Your changes will not be saved.

#ifndef vtkArrayDispatchArrayList_h
#define vtkArrayDispatchArrayList_h

#include "vtkTypeList.h"
#include "vtkAOSDataArrayTemplate.h"
#include "vtkStructuredPointArray.h"

namespace vtkArrayDispatch {
VTK_ABI_NAMESPACE_BEGIN

typedef vtkTypeList::Unique<
  vtkTypeList::Create<
    vtkAOSDataArrayTemplate<char>,
    vtkAOSDataArrayTemplate<double>,
    vtkAOSDataArrayTemplate<float>,
    vtkAOSDataArrayTemplate<int>,
    vtkAOSDataArrayTemplate<long>,
    vtkAOSDataArrayTemplate<long long>,
    vtkAOSDataArrayTemplate<short>,
    vtkAOSDataArrayTemplate<signed char>,
    vtkAOSDataArrayTemplate<unsigned char>,
    vtkAOSDataArrayTemplate<unsigned int>,
    vtkAOSDataArrayTemplate<unsigned long>,
    vtkAOSDataArrayTemplate<unsigned long long>,
    vtkAOSDataArrayTemplate<unsigned short>,
    vtkAOSDataArrayTemplate<vtkIdType>
  >
>::Result Arrays;

typedef vtkTypeList::Unique<
  vtkTypeList::Create<

    vtkStructuredPointArray<double>
  >
>::Result ReadOnlyArrays;

typedef vtkTypeList::Unique<
vtkTypeList::Append<Arrays,
ReadOnlyArrays>::Result
>::Result AllArrays;

VTK_ABI_NAMESPACE_END

} // end namespace vtkArrayDispatch
#endif // vtkArrayDispatchArrayList_h

