// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause
#ifndef vtkVersionFull_h
#define vtkVersionFull_h

/* This is in its own header to reduce build dependencies */

#define VTK_VERSION_FULL ""

#endif
